// Copyright 1998-2001 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OIF/OIAnimationInstance.h>

#import <Foundation/Foundation.h>
#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>

#import <OIF/OIAnimation.h>
#import <OIF/OIAnimationFrame.h>

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Frameworks/OIF/OIAnimationInstance.m,v 1.25 2001/02/15 15:11:48 kc Exp $")

@interface OIAnimationInstance (Private)
- (void)displayNextFrame;
- (void)scheduledDisplayNextFrame;
- (void)cancelScheduledEvent;
@end

@implementation OIAnimationInstance

static OFScheduler *animationScheduler;

+ (void)initialize;
{
    static BOOL initialized = NO;

    [super initialize];
    if (initialized)
        return;
    initialized = YES;

    animationScheduler = [[OFScheduler mainScheduler] subscheduler];
    [animationScheduler retain];
}

- initWithAnimation:(OIAnimation *)anAnimation;
{
    if (![super init])
        return nil;

    OFWeakRetainConcreteImplementation_INIT;
    animation = [anAnimation retain];
    remainingLoops = [animation loopCount];
    nextFrameEventLock = [[NSLock alloc] init];
    
    return self;
}

- (void)dealloc;
{
    OFWeakRetainConcreteImplementation_DEALLOC;
    OBASSERT(frame == nil); // Our frame was already released by -invalidateWeakRetains
    [animation release];
    [nextFrameEventLock release];
    [nextFrameEvent release]; // Assert this is nil?
    
    [super dealloc];
}

// Called by the animation, could possibly be in another thread.

- (void)animationEnded;
{
    if (remainingLoops == 0 || nextFrame <= 1)
        return;

    if (--remainingLoops > 0 && nextFrame != NSNotFound) {
        nextFrame = 0;
        [self displayNextFrame];
    }
}

- (void)animationReceivedFrame:(OIAnimationFrame *)aFrame;
{
    if (nextFrame == NSNotFound)
        return;

    nextFrame++;
    if (frame != aFrame) {
        [frame removeObserver:self];
        [frame release];
        frame = [aFrame retain];
        if (!haveSize && [frame hasSize])
            [self setSize:[frame size]];
        [self updateImage:[frame image]];
        [frame addObserver:self];
    }

    if ([self observerCount] != 0) {
        OFScheduledEvent *event;

        event = [animationScheduler scheduleSelector:@selector(scheduledDisplayNextFrame) onObject:self withObject:nil afterTime:[frame delayInterval]];
        [nextFrameEventLock lock];
        OBASSERT(nextFrameEvent == nil);
        nextFrameEvent = [event retain];
        [nextFrameEventLock unlock];
    }
}

OFWeakRetainConcreteImplementation_IMPLEMENTATION

- (void)invalidateWeakRetains;
{
    [frame removeObserver:self];
    [frame release];
    frame = nil;
}

// OIImage subclass

- (void)startAnimation;
{
    [self cancelScheduledEvent];
    remainingLoops = [animation loopCount];
    nextFrame = 0;
    [self displayNextFrame];
}

- (void)stopAnimation;
{
    [self cancelScheduledEvent];
    nextFrame = NSNotFound;
}


- (void)addObserver:(id <OIImageObserver, OFWeakRetain>)anObserver;
{
    BOOL jumpstartAnimation;

    jumpstartAnimation = [self observerCount] == 0;
    [super addObserver:anObserver];
    if (jumpstartAnimation) {
        // We've added our first observer, time to animate
        [self startAnimation];
    }
}

- (void)removeObserver:(id <OIImageObserver, OFWeakRetain>)anObserver;
{
    [super removeObserver:anObserver];
    if ([self observerCount] == 0) {
        // We've removed our last observer, no need to animate
        [self stopAnimation];
    }
}

// OIImageObserver protocol

- (void)imageDidSize:(OIImage *)anOmniImage;
{
    if (!haveSize)
        [self setSize:[anOmniImage size]];
}

- (void)imageDidUpdate:(OIImage *)anOmniImage;
{
    [self updateImage:[anOmniImage image]];
}

- (void)imageDidAbort:(OIImage *)anOmniImage;
{
    [self cancelScheduledEvent];
    // Pass the abort along to our observers
    [self abortImage];
}

// NSObject subclass (Debugging)

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *debugDictionary;

    debugDictionary = [super debugDictionary];
    [debugDictionary setObject:[NSNumber numberWithInt:remainingLoops] forKey:@"remainingLoops"];
    [debugDictionary setObject:[NSNumber numberWithInt:nextFrame] forKey:@"nextFrame"];

    return debugDictionary;
}

@end

@implementation OIAnimationInstance (Private)

- (void)displayNextFrame;
{
    if (nextFrame == NSNotFound)
        return;

    [animation animationInstance:self wantsFrame:nextFrame];
}

- (void)scheduledDisplayNextFrame;
{
    [nextFrameEventLock lock];
    if (nextFrameEvent != nil) {
        // We're done with the event that just called us
        [nextFrameEvent release];
        nextFrameEvent = nil;
    }
    [nextFrameEventLock unlock];
    [self displayNextFrame];
}

- (void)cancelScheduledEvent;
{
    OFScheduledEvent *event;

    [nextFrameEventLock lock];
    event = nextFrameEvent; // Inherit retain
    nextFrameEvent = nil;
    [nextFrameEventLock unlock];
    if (event != nil) {
        [animationScheduler abortEvent:event];
        [event release];
    }
}

@end
